@@ -30,6 +30,8 @@
 # OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 # SOFTWARE.
 
+import asyncio
+import contextlib
 from typing import Dict, Set
 
 from libqtile import utils
@@ -87,6 +89,24 @@ class Subscribe:
         """
         return self._subscribe("startup_complete", func)
 
+    def shutdown(self, func):
+        """Called before qtile is shutdown
+
+        **Arguments**
+
+        None
+        """
+        return self._subscribe("shutdown", func)
+
+    def restart(self, func):
+        """Called before qtile is restarted
+
+        **Arguments**
+
+        None
+        """
+        return self._subscribe("restart", func)
+
     def setgroup(self, func):
         """Called when group is changed
 
@@ -124,7 +144,8 @@ class Subscribe:
         return self._subscribe("changegroup", func)
 
     def focus_change(self, func):
-        """Called when focus is changed
+        """Called when focus is changed, including moving focus between groups or when
+        focus is lost completely
 
         **Arguments**
 
@@ -146,22 +167,11 @@ class Subscribe:
 
         **Arguments**
 
-        None
+            * ``Group`` receiving the new window
+            * ``Window`` added to the group
         """
         return self._subscribe("group_window_add", func)
 
-    def window_name_change(self, func):
-        """Called whenever a windows name changes
-
-        Deprecated: use client_name_updated
-        **Arguments**
-
-        None
-        """
-        msg = 'window_name_change is deprecated, use client_name_updated instead'
-        logger.warning(msg, DeprecationWarning)
-        return self._subscribe("window_name_change", func)
-
     def client_new(self, func):
         """Called before Qtile starts managing a new client
 
@@ -170,7 +180,7 @@ class Subscribe:
 
         **Arguments**
 
-            * ``window.Window`` object
+            * ``Window`` object
 
         Examples
         --------
@@ -194,7 +204,7 @@ class Subscribe:
 
         **Arguments**
 
-            * ``window.Window`` object of the managed window
+            * ``Window`` object of the managed window
         """
         return self._subscribe("client_managed", func)
 
@@ -203,16 +213,16 @@ class Subscribe:
 
         **Arguments**
 
-            * ``window.Window`` object of the killed window.
+            * ``Window`` object of the killed window.
         """
         return self._subscribe("client_killed", func)
 
     def client_focus(self, func):
-        """Called whenever focus changes
+        """Called whenever focus moves to a client window
 
         **Arguments**
 
-            * ``window.Window`` object of the new focus.
+            * ``Window`` object of the new focus.
         """
         return self._subscribe("client_focus", func)
 
@@ -221,7 +231,7 @@ class Subscribe:
 
         **Arguments**
 
-            * ``window.Window`` of window entered
+            * ``Window`` of window entered
         """
         return self._subscribe("client_mouse_enter", func)
 
@@ -230,7 +240,7 @@ class Subscribe:
 
         **Arguments**
 
-            * ``window.Window`` of client with updated name
+            * ``Window`` of client with updated name
         """
         return self._subscribe("client_name_updated", func)
 
@@ -239,7 +249,7 @@ class Subscribe:
 
         **Arguments**
 
-            * ``window.Window`` of client with hint change
+            * ``Window`` of client with hint change
         """
         return self._subscribe("client_urgent_hint_changed", func)
 
@@ -258,7 +268,7 @@ class Subscribe:
 
         **Arguments**
 
-            * ``window.Window`` of client with changed icon
+            * ``Window`` of client with changed icon
         """
         return self._subscribe("net_wm_icon_change", func)
 
@@ -285,23 +295,12 @@ class Subscribe:
         return self._subscribe("selection_change", func)
 
     def screen_change(self, func):
-        """Called when a screen is added or screen configuration is changed (via xrandr)
-
-        Common usage is simply to call ``qtile.cmd_restart()`` on each event
-        (to restart qtile when there is a new monitor):
+        """Called when the output configuration is changed (e.g. via randr in X11).
 
         **Arguments**
 
-            * ``xproto.randr.ScreenChangeNotify`` event
+            * ``xproto.randr.ScreenChangeNotify`` event (X11) or None (Wayland).
 
-        Examples
-        --------
-
-        ::
-
-            @libqtile.hook.subscribe.screen_change
-            def restart_on_randr(ev):
-                libqtile.qtile.cmd_restart()
         """
         return self._subscribe("screen_change", func)
 
@@ -355,6 +354,17 @@ class Unsubscribe(Subscribe):
 unsubscribe = Unsubscribe()
 
 
+def _fire_async_event(co):
+    loop = None
+    with contextlib.suppress(RuntimeError):
+        loop = asyncio.get_running_loop()
+
+    if loop is None:
+        asyncio.run(co)
+    else:
+        asyncio.ensure_future(co)
+
+
 def fire(event, *args, **kwargs):
     if event not in subscribe.hooks:
         raise utils.QtileError("Unknown event: %s" % event)
@@ -362,12 +372,11 @@ def fire(event, *args, **kwargs):
         logger.debug("Internal event: %s(%s, %s)", event, args, kwargs)
     for i in subscriptions.get(event, []):
         try:
-            i(*args, **kwargs)
+            if asyncio.iscoroutinefunction(i):
+                _fire_async_event(i(*args, **kwargs))
+            elif asyncio.iscoroutine(i):
+                _fire_async_event(i)
+            else:
+                i(*args, **kwargs)
         except:  # noqa: E722
             logger.exception("Error in hook %s", event)
-
-
-@subscribe.client_name_updated
-def _fire_window_name_change(window):
-    "This should eventually be removed"
-    fire("window_name_change")
